
#
# Copyright 2006 VMware, Inc.  All rights reserved.
#

use 5.006001;
use strict;
use warnings;

use Carp qw(confess croak);
use XML::LibXML;
use LWP::UserAgent;
use LWP::ConnCache;
use HTTP::Request;
use HTTP::Headers;
use HTTP::Response;
use HTTP::Cookies;
use Data::Dumper;


##################################################################################
package Util;

our $is_connected = 0;

sub check_fault {
   my $result = shift;
   if ($result->fault) {
      Carp::confess($result->fault);
   }   
   return $result->result;
}

# --------------------------------------------------------------------------------
# Description: Establish a connection to the Virtual Center or ESX server.
# Input: subroutine style:  Util::connect($url, $username, $password})
#        where
#           $url      - the URL of the server to which the client connects
#           $username - user account on the VC or ESX server
#           $password - password for the user account
# Output: a Vim object instance
# --------------------------------------------------------------------------------
sub connect {
   # url should be <protocol>://<hostname>/sdk
   my ($url, $username, $password) = @_;
   
   # bug 217340
   @_ = ();
   
   my $retval = undef;
   
   if (!defined($url)) {
      $url = Opts::get_option('url');
   }
   
   if (Opts::option_is_set('sessionfile')) {
      my $sessionfile = Opts::get_option('sessionfile');
      $retval = Vim::load_session(service_url => $url, session_file => $sessionfile);
   } elsif ((Opts::option_is_set('username') || defined($username)) && (Opts::option_is_set('password') || defined($password))) {
      if (!defined($username)) {
         $username = Opts::get_option('username');
      }
      
      if (!defined($password)) {
         $password = Opts::get_option('password');
      }
      
      eval {
         $retval = Vim::login(service_url => $url, user_name => $username, password => $password);
      };
      if ($@) {
         my $error = "Error connecting to server at '$url'";
         if ($@ =~ /Connection refused/) {
            die "$error: Connection refused\n";
         } elsif ($@ =~ /Bad hostname/) {
            die "$error: Bad hostname\n";
         } elsif ($@ =~ /Connect failed/) {
            die "$error: Perhaps host is not a Virtual Center or ESX server\n";
         } elsif ($@ =~ /Login failed/) {
            die "$error: Login failed due to a bad username or password\n";            
         } elsif (ref $@ eq "SoapFault" && defined $@->{fault_string}) {
            die "$error: " . $@->{fault_string} . "\n";            
         } else {
            die "$@\n";
         }
      }
   } elsif (Opts::option_is_set('passthroughauth')) {
      eval {
         require Win32::OLE;
      };
      if ($@) {
         die "Cannot load Win32::OLE: $@\n";
      }
      my $sspi = Win32::OLE->new("SSPICOM.SspiClientSession");
      if (!defined($sspi)) {
         die "Cannot create SSPI session\n";
      }
      my $package = Opts::get_option("passthroughauthpackage");
      if (!$sspi->Begin($package) || $sspi->GetState() ne "Negotiating") {
         die "SSPI Begin failed: package '$package'\n";
      }
      my $done = 0;
      while (!$done) {
         my $token = $sspi->GetToken();
         eval {
            $retval = Vim::login_by_sspi(service_url => $url, token => $token);
         };
         if ($@) {
            my $fault = $@;
            if (ref $fault eq "SoapFault" && defined $fault->{detail} &&
                ref $fault->{detail} eq "SSPIChallenge") {
               my $serverToken = $fault->{detail}->{base64Token};
               if (!$sspi->Initialize($serverToken) ||
                   $sspi->GetState() eq "Error") {
                  die "SSPI Initialize failed\n";
               }
            } else {
               $done = 1;
               my $error = "Error connecting to server at '$url'";
               die "$error: $@\n";
            }
         } else {
            $done = 1;
         }
      }
      $sspi->Reset();
   } else {
      die "ERROR: No session file or username/password provided\n";
   }
   
   $is_connected = 1;
   return $retval;
}

# --------------------------------------------------------------------------------
# Description: Disconnect a session with the Virtual Center or ESX server.
# Input: none
# Output: none
# --------------------------------------------------------------------------------
sub disconnect {
   if ($is_connected == 1) {
      if (Opts::option_is_set('sessionfile')) {
         Vim::clear_session();
      } else {
         if (!Opts::option_is_set('savesessionfile')) {            
            Vim::logout();
         }
      }
      $is_connected = 0;
   }
   return;
}

sub create_entity_info {
   my $objectContent = shift;
   my $propSet = $objectContent->propSet;
   my $name;
   my $parent;
   foreach (@$propSet) {
      if ($_->name eq "name") {
         $name = $_->val;
      } elsif ($_->name eq "parent") {
         $parent = $_->val;
      }
   }
   return { name => $name, parent => $parent };
}

# --------------------------------------------------------------------------------
# Description: Return the inventory path for the specified managed entity.  
# Input: subroutine style:  Util::get_inventory_path($entity, $vim})
#        where
#           $entity   - a managed entity view
#           $vim      - a Vim object instance
# Output: a string that identifies the inventory path of the managed entity.
# --------------------------------------------------------------------------------
sub get_inventory_path {
   # bug 193817
   my ($entity, $service) = @_;
   my $propertyCollector = $service->{service_content}->propertyCollector;

   # Retrieve the "name" and "parent" properties for all the managed entities
   # reachable by following "parent" properties starting from $entity.
   my $objectContents = Util::check_fault(
      $service->{vim_service}->RetrieveProperties(
      _this => $propertyCollector,
       specSet => [
        PropertyFilterSpec->new(
            propSet => [
          PropertySpec->new(
             type => "ManagedEntity",
             pathSet => [ "name", "parent" ]
          )
          ],
           objectSet => [
           ObjectSpec->new(
            obj => $entity->{mo_ref},
             selectSet => [
             TraversalSpec->new(
                name => "ParentTraversalSpec",
               type => "ManagedEntity",
               path => "parent",
               selectSet => [
                  SelectionSpec->new(name => "ParentTraversalSpec") 
               ]
            )
             ]
          )
           ]
        )
     ]
      )
   );

   # Create a hash containing an entry for each manaaged entity retrieved by
   # the RetrieveProperties call above.
   my %entityMap;
   foreach (@$objectContents) {
      my $objectContent = $_;
      if (defined $objectContent->missingSet) {
         die $objectContent->misingSet . "\n";
      }
      $entityMap{$objectContent->obj->value} = create_entity_info($objectContent);
   }

   # Use entityMap to walk the "parent" properties of the managed entities,
   # starting with managedEntity, computing the inventory path (from end to
   # beginning) along the way.
   my $current = $entityMap{$entity->{mo_ref}->value};
   my $parent = $current->{parent};
   if (! defined $parent) {
      return "";
   }
   my @path;
   while (1) {
      unshift(@path, $current->{name});
      $current = $entityMap{$parent->value};
      $parent = $current->{parent};
      if (! defined $parent) {
	 last;
      }
   }
   return join "/", @path
}

1;
##################################################################################


##################################################################################
package Vim;

use Data::Dumper;
use URI::Escape;

our $vim_global = undef;
our $server_version = undef;
our $vim_namespace = undef;

sub get_server_version { $server_version };
sub get_vim_namespace { $vim_namespace };

sub set_server_version { 
   my ($arg) = @_; 
   $server_version = $arg;
}

sub new {
   my ($class, %args) = @_;
   my $self = {};
   bless $self, $class;
   if (! defined $args{service_url}) {
      Carp::confess("service_url argument missing");
   }
   $self->{service_url} = $args{service_url};
   return $self;
}

# bug 231138
# --------------------------------------------------------------------------------
# Description: Return the current Vim object instance.
# Input:  none
# Output: $self       if called as vim->get_vim() 
#         $vim_global if called as Vim::get_vim()
# --------------------------------------------------------------------------------
sub get_vim {
   my $self = shift;
   
   if ($self and ref $self eq 'Vim') {
      return $self;
   } else {
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }
      return $vim_global;
   }
}

sub query_server_version {
   my $url = shift;
   if ($url =~ s|http(s?)://(.*)/sdk.*|http$1://$2/sdk/vimService.wsdl|i) {
      my $user_agent = LWP::UserAgent->new;
      my $cookie_jar = HTTP::Cookies->new(ignore_discard => 1);
      $user_agent->cookie_jar($cookie_jar);
      $user_agent->protocols_allowed(['http', 'https']);
  
      my $http_header = HTTP::Headers->new(Content_Type => 'text/xml');
      my $request = HTTP::Request->new('GET', $url);
     
      my $response = $user_agent->request($request);   
     
      if ($response->content =~ /Connection refused/) {
         die $response->content . "\n";
      } elsif ($response->content =~ /Bad hostname/) {
         die $response->content . "\n";
      } elsif ($response->content =~ /Connect failed/) {
         die $response->content . "\n";
      }
      
      my $xml_parser = XML::LibXML->new;   
      my $result;   

      eval { $result = $xml_parser->parse_string($response->content) };
      if ($@) {
         die "Error: Server version unavailable at '$url' $@\n";
      }   
   
      my $targetNamespace = $result->documentElement()->getAttribute('targetNamespace');
        
      my ($server_version) = $targetNamespace =~ 
                            /(?:urn:vim)([0-9a-zA-Z]+)(?:Service)/;
                      
      if ($server_version) {
         # TO BE REMOVED: kludge for development phase
         if ($server_version eq "uber") {
            $server_version = "25";
         }
         if ($server_version eq "dev") {
            $server_version = "25";
         }         
         return $server_version;
      } else {
         die "Error: Server version '$targetNamespace' at '$url' unsupported\n";
      }
   } else {
      die "Error: Invalid URL '$url'\n";
   }
   return $server_version;
}

# --------------------------------------------------------------------------------
# Description: Establish a session with the Virtual Center or ESX server.
# Input: subroutine style:  Vim::login(%{service_url, user_name, password})
#        method call style: vim->login(%{user_name, password})
#        where
#           service_url - the URL of the server to which the client connects
#           user_name   - user account on the VC or ESX server
#           password    - password for the user account
# Output: a Vim object instance
# --------------------------------------------------------------------------------
sub login {
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      # object reference 
      my ($self, %args) = @_;
      
      # bug 217340
      @_ = ();
   
      my $service_url = $self->{service_url};
      my $user_name = $args{user_name};
      my $password = $args{password};

      #
      # Load the appropriate Runtime/Stubs if not done yet, and
      # set the default namespace.
      #
      unless ($server_version) {
         $server_version = query_server_version($service_url);
      }
      
      unless ($vim_namespace) {
         $vim_namespace = "vim$server_version";
         eval qq(require VMware::VIM${server_version}Runtime);
         die "$@\n" if $@;
         eval qq(require VMware::VIM${server_version}Stub);
         die "$@\n" if $@;
      }
      
      my $si_mo_ref = ManagedObjectReference->new(type => 'ServiceInstance',
                                                  value => 'ServiceInstance');
      $self->{vim_service} = VimService->new($service_url);
      my $sc = Util::check_fault($self->{vim_service}->RetrieveServiceContent(_this => $si_mo_ref));
      my $result = $self->{vim_service}->Login(_this => $sc->sessionManager,
                                               userName => $user_name,
                                               password => $password);
      if ($result->fault) {
         die $result->fault->{fault_string} . "\n";
      }
      $self->{service_content} = $sc;
   } else {
      my %args = @_;
      
      # bug 217340
      @_ = ();
   
      my $service_url = $args{service_url};
      $vim_global = Vim->new(service_url => $service_url);      
      my $user_name = $args{user_name};
      my $password = $args{password};
      $vim_global->login(user_name => $user_name, password => $password);
   }
   return $vim_global;
}

# ----------------------------------------------------------------------------------
# Description: Establish a session with the Virtual Center or ESX server using SSPI.
# Input: subroutine style:  Vim::login_by_sspi(%{service_url, token})
#        method call style: vim->login_by_sspi(%{token})
#        where
#           service_url - the URL of the server to which the client connects
#           token       - the SSPI token
# Output: a Vim object instance
# ----------------------------------------------------------------------------------
sub login_by_sspi {
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      # object reference 
      my ($self, %args) = @_;
      my $service_url = $self->{service_url};
      my $token = $args{token};
      
      #
      # Load the appropriate Runtime/Stubs if not done yet, and
      # set the default namespace.
      #
      unless ($server_version) {
         $server_version = query_server_version($service_url);
      }
      
      unless ($vim_namespace) {
         $vim_namespace = "vim$server_version";
         eval qq(require VMware::VIM${server_version}Runtime);
         die "$@\n" if $@;
         eval qq(require VMware::VIM${server_version}Stub);
         die "$@\n" if $@;
      }
      
      if (!defined $self->{vim_service}) {           
         my $si_moref = ManagedObjectReference->new(type => 'ServiceInstance',
                                                    value => 'ServiceInstance');
         $self->{vim_service} = VimService->new($service_url);
         my $sc = Util::check_fault($self->{vim_service}->RetrieveServiceContent(_this => $si_moref));
         $self->{service_content} = $sc;
      }
      Util::check_fault($self->{vim_service}->LoginBySSPI(
                  _this => $self->{service_content}->sessionManager,
                  base64Token => $token));
   } else {
      my %args = @_;
      if (!defined($vim_global)) {
         my $service_url = $args{service_url};
         $vim_global = Vim->new(service_url => $service_url);
      }      
      my $token = $args{token};
      $vim_global->login_by_sspi(token => $token);
   }
   return $vim_global;
}

# --------------------------------------------------------------------------------
# Description: Retrieve the session id corresponding to the current session.
# Input:  none
# Output: a session id for use by load_session()
# --------------------------------------------------------------------------------
sub get_session_id {
   my $self;
   my %args;
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
   } else {
      %args = @_;
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }
   }
   my $id = $self->{vim_service}->get_session_id();
   return $id;
}

# --------------------------------------------------------------------------------
# Description: Save a session cookie to a specified file.
# Input: subroutine style:  Vim::save_session(%{session_file})
#        method call style: vim->save_session(%{session_file})
#        where
#           session_file - full path file name where session is saved to
# Output: none
# --------------------------------------------------------------------------------
sub save_session {
   my $self;
   my %args;
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
   } else {
      %args = @_;
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }
   }
   if (!defined($args{session_file})) {
         Carp::croak("Missing session_file parameter\n");
   }
   $self->{vim_service}->save_session($args{session_file});
   return;
}

# --------------------------------------------------------------------------------
# Description: Load a session cookie from a specified file.
# Input: subroutine style:  Vim::load_session(%{service_url, session_file})
#        where
#           service_url - the URL of the server to which the client connects
#           session_file - full path file name where session is saved to
# Output: none
# --------------------------------------------------------------------------------
sub load_session {
   my $self;
   my %args;
   my $url;
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
      return $self;
   } else {
      %args = @_;
      
      # set server using data from session file
      $url = $args{service_url};
      my $file = $args{session_file};
      if (defined($file)) {
         local *SFILE;
         my $line;
             
         open(SFILE, $file) or die "ERROR: Can't open session file \'$file\'\n";
         $line = <SFILE>;
         $line = <SFILE>;
         if (defined($line) && ($line =~ /\w*\s*domain=(.*).local/)) {
            Opts::set_option('server' => $1);
            $url = Opts::get_option('url');
         } else {
            if (defined($line) && ($line =~ /\w*\s*domain=(.*?);/)) {
               Opts::set_option('server' => $1);
               $url = Opts::get_option('url');
            }
         }
         close(SFILE);
      }    
      $vim_global = Vim->new(service_url => $url);
      $self = $vim_global;
   }
   
   #
   # Load the appropriate Runtime/Stubs if not done yet, and
   # set the default namespace.
   #
   unless ($server_version) {
      $server_version = query_server_version($url);
   }
      
   unless ($vim_namespace) {
      $vim_namespace = "vim$server_version";
      eval qq(require VMware::VIM${server_version}Runtime);
      die $@ if $@;
      eval qq(require VMware::VIM${server_version}Stub);
      die $@ if $@;
   }
      
   my $si_mo_ref = ManagedObjectReference->new(type => 'ServiceInstance',
                                               value => 'ServiceInstance');
   $self->{vim_service} = VimService->new($self->{service_url});
   if ($args{session_id}) {
      $self->{vim_service}->load_session_id($args{session_id});
   }
   else {
      $self->{vim_service}->load_session($args{session_file});
   }
   my $sc = Util::check_fault($self->{vim_service}->RetrieveServiceContent(_this => $si_mo_ref));
   $self->{service_content} = $sc;
   return $vim_global;
}

# --------------------------------------------------------------------------------
# Description: Terminate the current session loaded by load_session().
# Input:  none
# Output: none
# --------------------------------------------------------------------------------
sub clear_session {
   my $self = shift;
   unless ($self && ref $self eq 'Vim') {
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }
      $self = $vim_global;
   }
   $self->{vim_service}->unload_session();
   return;
}

# --------------------------------------------------------------------------------
# Description: Return the current VimService object instance.
# Input:  none
# Output: a VimService object instance
# --------------------------------------------------------------------------------
sub get_vim_service {
   my $self = shift;
   unless ($self && ref $self eq 'Vim') {
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }
      $self = $vim_global;
   }
   
   if (defined $self->{vim_service}) {
      return $self->{vim_service};
   } else {
      Carp::confess("Vim uninitialized\n");
      return undef;
   }
}

# --------------------------------------------------------------------------------
# Description: Return a ServiceInstance object instance.
# Input:  none
# Output: a ServiceInstance object instance
# --------------------------------------------------------------------------------
sub get_service_instance {
   my $self = shift;
   unless ($self && ref $self eq 'Vim') {
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }
      $self = $vim_global;
   }
   my $si_mo_ref = ManagedObjectReference->new(type  => 'ServiceInstance',
                                               value => 'ServiceInstance');
   return $self->get_view(mo_ref => $si_mo_ref);
}

# --------------------------------------------------------------------------------
# Description: Return a ServiceContent object instance.
# Input:  none
# Output: a ServiceContent object instance
# --------------------------------------------------------------------------------
sub get_service_content {
   my $self;
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      $self = shift;
   } else {
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }      
   }
   if (defined $self->{service_content}) {
      return $self->{service_content};
   } else {
      Carp::confess("Vim uninitialized");
      return undef;
   }
}

sub match {
   my ($pat, $str) = @_;
   
   # bug 178470
   $str = uri_unescape($str);
   
   if (!ref($pat)) {
      return $str eq $pat;
   } elsif (ref($pat) ne 'Regexp') {
      die "Regexp or string arg expected\n";
   }
   return $str =~ /$pat/;
}

sub mismatch {
   my ($pat, $str) = @_;
   
   # bug 178470
   $str = uri_unescape($str);
   
   if (!ref($pat)) {
      return $str ne $pat;
   } elsif (ref($pat) ne 'Regexp') {
      die "Regexp or string arg expected\n";
   }
   return $str !~ /$pat/;
}

# ----------------------------------------------------------------------------------
# Description: Search the inventory tree for 1st managed entity that matches the 
#              specified entity type.  The search begins with the root folder 
#              unless the begin_entity parameter is specified.
# Input: subroutine style:  Vim::find_entity_view(%{view_type, begin_entity, filter})
#        method call style: vim->find_entity_view(%{view_type, begin_entity, filter})
#        where
#           view_type     - Folder | HostSystem | ResourcePool | VirtualMachine |
#                           ComputeResource | DataCenter | ClusterComputeResource
#           begin_entity  - starting managed object reference
#           filter        - a hash of one or more name value pair
# Output: a managed object view instance
# ----------------------------------------------------------------------------------
sub find_entity_view {
   my $self = undef;
   my $sc;
   my $service;
   my %args;
 
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
   } else {
      %args = @_;
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }      
   }
   $service = $self->{vim_service};
   $sc = $self->{service_content};

   if (! exists($args{view_type})) {
      Carp::confess('view_type argument is required');
   }
   my $view_type = $args{view_type};
   delete $args{view_type};
   if (! $view_type->isa('EntityViewBase')) {
      Carp::confess("$view_type is not a ManagedEntity");
   }
   my $properties = "";
   if (exists ($args{properties})) {
      if (defined($args{properties})) {
         $properties = $args{properties};
      }
      delete $args{properties};
   }
   my $begin_entity = $sc->rootFolder;
   if (exists ($args{begin_entity})) {
      $begin_entity = $args{begin_entity};
      delete $args{begin_entity};
   }   
   my $filter = {};
   if (exists ($args{filter})) {
      $filter = $args{filter};
      delete $args{filter};
   }
   
   my @remaining = keys %args;
   if (@remaining > 0) {
      Carp::confess("Unexpected argument @remaining");
   }
   my %filter_hash;
   foreach (keys %$filter) {
      my $key = $_;
      my $keyvalue = $filter->{$_};
      my $index = index($_, "-");
      if($index == 0) {
         $key = substr($key,1);
      }
      $filter_hash{$key} = $keyvalue;
   }
   my $property_spec = PropertySpec->new(all => 0,
                                         type => $view_type->get_backing_type(),
                                         pathSet => [keys %filter_hash]);
   my $property_filter_spec = 
      $view_type->get_search_filter_spec($begin_entity, [$property_spec]);
   my $obj_contents =
      $service->RetrieveProperties(_this => $sc->propertyCollector,
                                   specSet => $property_filter_spec);   
   my $result = Util::check_fault($obj_contents);   
   my $filtered_obj;
   foreach (@$result) {
      my $obj_content = $_;
      if (keys %$filter == 0) {
         $filtered_obj = $obj_content->obj;
         last;
      }

      my %prop_hash;    
      my $prop_set = $obj_content->propSet;
      if (! $prop_set) {
         next;
      }
      foreach (@$prop_set) {
         $prop_hash{$_->name} = $_->val;
      }
      my $matched = 1;
      foreach (keys %$filter) {
         my $index = index($_, "-");
         my $regex = $filter->{$_};
         my $flag = 0;
         if($index == 0) {
            $_ = substr($_,1);
            $flag=1;
         }
         if (exists ($prop_hash{$_})) {         
            my $val;
            if (ref $prop_hash{$_}) {
               my $class = ref $prop_hash{$_};
               if ($class->isa("SimpleType")) {
                  $val = $prop_hash{$_}->val();
               } else {
                  Carp::croak("Filtering is only supported for Simple Types\n");
               }
            } else {
               $val = $prop_hash{$_};
            }
            
            if($flag == 0) {
               if (not match($regex,$val)) {
                  $matched = 0;
                  last;
               }
            }
            else {
               if (not mismatch($regex,$val)) {
                  $matched = 0;
                  last;
               }
            }
         }
      }
      if ($matched) {
         $filtered_obj = $obj_content->obj;
      }
   }
   if (! $filtered_obj) {
      return undef;
   } else {
      return $self->get_view(mo_ref => $filtered_obj,
                             view_type => $view_type,
                             properties => $properties);
   }
}

# ----------------------------------------------------------------------------------
# Description: Search the inventory tree for all managed entity that matches the 
#              specified entity type.  The search begins with the root folder 
#              unless the begin_entity parameter is specified.
# Input: subroutine style:  Vim::find_entity_views(%{view_type, begin_entity, filter})
#        method call style: vim->find_entity_views(%{view_type, begin_entity, filter})
#        where
#           view_type     - Folder | HostSystem | ResourcePool | VirtualMachine |
#                           ComputeResource | DataCenter | ClusterComputeResource
#           begin_entity  - starting managed object reference
#           filter        - a hash of one or more name value pair
# Output: a reference to an array managed object view instances
# ----------------------------------------------------------------------------------
sub find_entity_views {
   my $self = undef;
   my $sc;
   my $service;
   my %args;
    
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
   } else {
      %args = @_;
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }      
   }
   $service = $self->{vim_service};
   $sc = $self->{service_content};   
   
   if (! exists($args{view_type})) {
      Carp::confess('view_type argument is required');
   }
   my $view_type = $args{view_type};
   delete $args{view_type};
   if (! $view_type->isa('EntityViewBase')) {
      Carp::confess("$view_type is not a ManagedEntity");
   }
   my $begin_entity = $sc->rootFolder;
   if (exists ($args{begin_entity})) {
      $begin_entity = $args{begin_entity};
      delete $args{begin_entity};
   }
   my $filter = {};
   if (exists ($args{filter})) {
      $filter = $args{filter};
      delete $args{filter};
   }
   my $properties = "";
   if (exists ($args{properties})) {
      if (defined($args{properties})) {
         $properties = $args{properties};
      }
      delete $args{properties};
   }
   my @remaining = keys %args;
   if (@remaining > 0) {
      Carp::confess("Unexpected argument @remaining");
   }
   
   my %filter_hash;
   foreach (keys %$filter) {
      my $key = $_;
      my $keyvalue = $filter->{$_};
      my $index = index($_, "-");
      if($index == 0) {
         $key = substr($key,1);
      }
      $filter_hash{$key} = $keyvalue;
   }
   
   my $property_spec = PropertySpec->new(all => 0,
                                         type => $view_type->get_backing_type(),
                                         pathSet => [keys %filter_hash]);
   my $property_filter_spec = 
      $view_type->get_search_filter_spec($begin_entity, [$property_spec]);
   my $obj_contents =
      $service->RetrieveProperties(_this => $sc->propertyCollector,
                                   specSet => $property_filter_spec);   
   my $result = Util::check_fault($obj_contents);   
   my @filtered_objs;
   foreach (@$result) {
      my $obj_content = $_;
      if (keys %$filter == 0) {
         push @filtered_objs, $obj_content->obj;
         next;
      }

      my %prop_hash;    
      my $prop_set = $obj_content->propSet;
      if (! $prop_set) {
         next;
      }
      foreach (@$prop_set) {
         $prop_hash{$_->name} = $_->val;
      }
      my $matched = 1;
      foreach (keys %$filter) {
         my $index = index($_, "-");
         my $regex = $filter->{$_};
         my $flag = 0;
         if($index == 0) {
            $_ = substr($_,1);
            $flag=1;
         }
         if (exists ($prop_hash{$_})) {
            my $val;
            if (ref $prop_hash{$_}) {
               my $class = ref $prop_hash{$_};
               if ($class->isa("SimpleType")) {
                  $val = $prop_hash{$_}->val();
               } else {
                  Carp::croak("Filtering is only supported for Simple Type\n");
               }
            } else {
               $val = $prop_hash{$_};
            }
	    
            if($flag == 0) {
               if (not match($regex,$val)) {
                  $matched = 0;
                  last;
               }
            }
            else {
               if (not mismatch($regex,$val)) {
                  $matched = 0;
                  last;
               }
            }
         }
      }
      if ($matched) {
         push @filtered_objs, $obj_content->obj;
      }
   }
   if (! @filtered_objs) {
      return [];
   } else {
      return $self->get_views(mo_ref_array => \@filtered_objs,
                              view_type => $view_type,
                              properties => $properties);
   }   
}

# ----------------------------------------------------------------------------------
# Description: Retrieve the properties of a single managed object. 
# Input: subroutine style:  Vim::get_view(%{mo_ref, view_type})
#        method call style: vim->get_view(%{mo_ref, view_type})
#        where
#           mo_ref     - a managed object reference
#           view_type  - the type of view to construct from the managed object
# Output: a managed object view instance
# ----------------------------------------------------------------------------------
sub get_view {
   my %args;   
   my $self = undef;
   my $service;
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
   } else {
      %args = @_;
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }      
   }
   $service = $self->{vim_service};
   
   if (! exists($args{mo_ref})) {
      Carp::confess("mo_ref argument is required");
   }
   my $mo_ref = $args{mo_ref};
   my $view_type = $mo_ref->type;
   if (exists ($args{view_type})) {
      $view_type = $args{view_type};
   }
   my $properties = "";
   if (exists($args{properties}) && 
       defined($args{properties}) && 
       $args{properties} ne "") {
      $properties = $args{properties};
   }
   my $view = $view_type->new($mo_ref, $self);
   $view->update_view_data($properties);
   return $view;
}

# ----------------------------------------------------------------------------------
# Description: Retrieve the properties of a set of managed object. 
# Input: subroutine style:  Vim::get_views(%{mo_ref_array, view_type})
#        method call style: vim->get_views(%{mo_ref_array, view_type})
#        where
#           mo_ref_array  - a reference to an array of managed object references
#           view_type     - the type of view to construct from the managed objects
# Output: a reference to an array managed object view instances
# ----------------------------------------------------------------------------------
sub get_views {
   my %args;   
   my $self = undef;
   my $service;
   my $sc;
   my ($first_arg) = @_;
   if ($first_arg && ref $first_arg eq 'Vim') {
      ($self, %args) = @_;
   } else {
      %args = @_;
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }     
   }
   $service = $self->{vim_service};
   $sc = $self->{service_content};

   if (! exists($args{mo_ref_array})) {
      Carp::confess("mo_ref_array argument is required");
   }
   my $mo_refs = $args{mo_ref_array};
   if (!defined($mo_refs) ||(@$mo_refs == 0)) {
      # Map an empty list of mo_refs to an empty list of views
      return [];
   }
   my $mo_ref = @$mo_refs[0];
   my $view_type = $mo_ref->type;
   if (exists ($args{view_type})) {
      $view_type = $args{view_type};
   }
     
   my $property_spec = $view_type->get_property_spec();
   my @object_specs;
   foreach (@$mo_refs) {
      push @object_specs, ObjectSpec->new(obj => $_);
   }
   my $prop_filter_spec = PropertyFilterSpec->new(propSet => $property_spec,
                                                  objectSet => \@object_specs);

   my $propertyCollector = $sc->propertyCollector;

   my $properties = "";
   if (exists($args{properties}) && 
       defined($args{properties}) && 
       $args{properties} ne "") {
      $properties = $args{properties};
      my $ptr = $prop_filter_spec->propSet;
      my $obj = $$ptr[0];
      $obj->all(0);  
      $obj->pathSet ($args{properties});
   }

   my $obj_contents =
      $service->RetrieveProperties(_this => $propertyCollector,
                                   specSet => $prop_filter_spec);
   Util::check_fault($obj_contents);
   my @views;
   foreach (@{$obj_contents->result}) {
      my $view = $view_type->new($_->obj, $self);
      $view->set_view_data($_, $properties);
      push @views, $view;
   }
   return \@views;
}

# --------------------------------------------------------------------------------
# Description: Close the connection to the server.
# Input:  none
# Output: none
# --------------------------------------------------------------------------------
sub logout {
   my ($first_arg) = @_;
   my $self;
   if ($first_arg && ref $first_arg eq 'Vim') {
      $self = $first_arg;
   } else {
      $self = $vim_global;
      unless (defined $vim_global) {
         Carp::croak("No global session in existence - perhaps need to login first\n");
      }      
   }   
   $self->{vim_service}->Logout(_this => $self->{service_content}->sessionManager);
   return;
}

1;
##################################################################################



##################################################################################
package ViewBase;

sub new {
   my ($class, $mo_ref, $vim) = @_;
   if (! $mo_ref) {
      Carp::confess("Missing required mo_ref argument");
   }
   my $self = { mo_ref => $mo_ref,
                vim => $vim };
   bless $self, $class;
}

# bug 226075
sub serialize {
   my ($self, $tag, $show_type) = @_;
   my $mo_ref = $self->{mo_ref};
   if($show_type) {
      return $mo_ref->serialize($tag, $show_type);   
   } else {
      return $mo_ref->serialize($tag);
   }
}

# bug 231138
sub get_vim {
   return shift->{vim};
}

sub get_backing_type {
   my $class = shift;
   return $class;
}

sub get_property_spec {
   my $class = shift;
   return [PropertySpec->new(all => 1,
                             type => $class)];
}

sub get_property_filter_spec {
   my ($class, $mo_ref) = @_;
   my $property_spec = $class->get_property_spec();
   my $object_spec = ObjectSpec->new(obj => $mo_ref);
   return PropertyFilterSpec->new(propSet => $property_spec,
                                  objectSet => [ $object_spec ]);   
}

sub update_view_data {
   my $self = shift;
   my $properties = shift;
   my $service = $self->{vim}->get_vim_service();
   my $property_filter_spec = (ref $self)->get_property_filter_spec($self->{mo_ref});
   my $propertyCollector = $self->{vim}->get_service_content()->propertyCollector;
   if (defined($properties) && $properties ne "") {
      my $ptr = $property_filter_spec->propSet;
      my $obj = $$ptr[0];
      $obj->all(0);  
      $obj->pathSet ($properties);
   }
   my $obj_contents =
      $service->RetrieveProperties(_this => $propertyCollector,
                                   specSet => $property_filter_spec);
   Util::check_fault($obj_contents);
   foreach (@{$obj_contents->result}) {
      $self->set_view_data($_, $properties);
   }
}

sub set_view_data {
   my ($self, $obj_content, $properties) = @_;
   my $prop_set = $obj_content->propSet;
   foreach (@$prop_set) {
      my $name = $_->name;
      if (!$properties || $properties eq "") {
         my @path_elements = split /\./, $name;
         $name = pop @path_elements;
      }
      $self->{$name} = $_->val;
   }
}


sub invoke {
   my ($self, $method, %args) = @_;
   my $runtime = $self->{vim}->get_vim_service();
   my $mo_ref = $self->{mo_ref};
   if (! $mo_ref) {
      Carp::confess("Required mo_ref argument not found");
   }
   $runtime->$method(_this => $mo_ref, %args);
}

sub get_property {
   my ($self, $path) = @_;
   my @subpaths = ();
   my $val;
   while ($path) {
      if (exists $self->{$path}) {
         $val = $self->{$path};
         last;
      } elsif ($path =~ /^(.+)\.([^.]+)$/) {
         $path = $1;
         unshift @subpaths, $2;
      }
   }
   if (defined($val)) {
      foreach (@subpaths) {
         $val = $val->{$_};
      }
   }
   return $val;
}

# bug 177573
sub waitForTask {
   my ($self, $task_ref, $progesscallbackfunc) = @_;
   my $task_view = $self->{vim}->get_view(mo_ref => $task_ref);
   my $progress = -1;
   while (1) {
      my $info = $task_view->info;
      if ($info->state->val eq 'success') {
         return $info->result;
      } elsif ($info->state->val eq 'error') {
         my $soap_fault = SoapFault->new;
         $soap_fault->name($info->error->fault);
         $soap_fault->detail($info->error->fault);
         $soap_fault->fault_string($info->error->localizedMessage);
         die "$soap_fault\n";
      }
      if(defined $progesscallbackfunc) {
         if(defined $info->progress && $info->progress != $progress) {
            &$progesscallbackfunc($info->progress);
            $progress = $info->progress;
         }
      }
      sleep 2;
      $task_view->update_view_data();
   }
}

1;
##################################################################################


##################################################################################
package EntityViewBase;
our @ISA = qw(ViewBase);

sub get_search_filter_spec {   
   my ($class, $mo_ref, $property_spec) = @_;
   my $resourcePoolTraversalSpec =
      TraversalSpec->new(name => 'resourcePoolTraversalSpec',
                         type => 'ResourcePool',
                         path => 'resourcePool',
                         skip => 0,
                         selectSet => [SelectionSpec->new(name => 'resourcePoolTraversalSpec'),
                                       SelectionSpec->new(name => 'resourcePoolVmTraversalSpec')]);      

   my $resourcePoolVmTraversalSpec =
      TraversalSpec->new(name => 'resourcePoolVmTraversalSpec',
                         type => 'ResourcePool',
                         path => 'vm',
                         skip => 0);   

   my $computeResourceRpTraversalSpec =
      TraversalSpec->new(name => 'computeResourceRpTraversalSpec',
                         type => 'ComputeResource',
                         path => 'resourcePool',
                         skip => 0,
                         selectSet => [SelectionSpec->new(name => 'resourcePoolTraversalSpec'),
                                       SelectionSpec->new(name => 'resourcePoolVmTraversalSpec')]);      
      
   
   my $computeResourceHostTraversalSpec =
      TraversalSpec->new(name => 'computeResourceHostTraversalSpec',
                         type => 'ComputeResource',
                         path => 'host',
                         skip => 0);
         
   my $datacenterHostTraversalSpec =
      TraversalSpec->new(name => 'datacenterHostTraversalSpec',
                         type => 'Datacenter',
                         path => 'hostFolder',
                         skip => 0,
                         selectSet => [SelectionSpec->new(name => "folderTraversalSpec")]);
   
   my $datacenterVmTraversalSpec =
      TraversalSpec->new(name => 'datacenterVmTraversalSpec',
                         type => 'Datacenter',
                         path => 'vmFolder',
                         skip => 0,
                         selectSet => [SelectionSpec->new(name => "folderTraversalSpec")]);

   my $hostVmTraversalSpec =
      TraversalSpec->new(name => 'hostVmTraversalSpec',
                         type => 'HostSystem',
                         path => 'vm',
                         skip => 0,
                         selectSet => [SelectionSpec->new(name => "folderTraversalSpec")]);
      
   my $folderTraversalSpec =
      TraversalSpec->new(name => 'folderTraversalSpec',
                         type => 'Folder',
                         path => 'childEntity',
                         skip => 0,
                         selectSet => [SelectionSpec->new(name => 'folderTraversalSpec'),
                                       SelectionSpec->new(name => 'datacenterHostTraversalSpec'),
                                       SelectionSpec->new(name => 'datacenterVmTraversalSpec',),
                                       SelectionSpec->new(name => 'computeResourceRpTraversalSpec'),
                                       SelectionSpec->new(name => 'computeResourceHostTraversalSpec'),
                                       SelectionSpec->new(name => 'hostVmTraversalSpec'),
                                       SelectionSpec->new(name => 'resourcePoolVmTraversalSpec'),
                                       ]);
   
   my $obj_spec = ObjectSpec->new(obj => $mo_ref,
                                  skip => 0,
                                  selectSet => [$folderTraversalSpec,
                                                $datacenterVmTraversalSpec,
                                                $datacenterHostTraversalSpec,
                                                $computeResourceHostTraversalSpec,
                                                $computeResourceRpTraversalSpec,
                                                $resourcePoolTraversalSpec,
                                                $hostVmTraversalSpec,
                                                $resourcePoolVmTraversalSpec]);
   
   return PropertyFilterSpec->new(propSet => $property_spec,
                                  objectSet => [$obj_spec]);
}

sub get_task_collector_view {
   my ($self, $recursion_option, %args) = @_;
   my $task_filter_spec = TaskFilterSpec->new;
   if (exists $args{filter}) {
      $task_filter_spec = $args{task_filter_spec};
   }
   $task_filter_spec->entity(
      TaskFilterSpecByEntity->new(entity => $self->{mo_ref},
                                  recursion => TaskFilterSpecRecursionOption->new($recursion_option)));
   my $task_mgr_view = $self->{vim}->get_view(mo_ref => $self->{vim}->get_service_content()->taskManager);
   my $task_collector = $task_mgr_view->CreateCollectorForTasks(filter => $task_filter_spec);
   return $self->{vim}->get_view(mo_ref => $task_collector);   
}

sub get_entity_only_tasks_collector_view {
   my ($self, %args) = @_;
   return $self->get_task_collector_view('self', %args);
}

sub get_all_tasks_view {
   my ($self, %args) = @_;
   return $self->get_task_collector_view('all', %args);   
}

sub get_event_collector_view {
   my ($self, $recursion_option, %args) = @_;
   my $event_filter_spec = EventFilterSpec->new;
   if (exists $args{filter}) {
      $event_filter_spec = $args{filter};
   }
   $event_filter_spec->entity(
      EventFilterSpecByEntity->new(entity => $self->{mo_ref},
                                   recursion => EventFilterSpecRecursionOption->new($recursion_option)));
   my $event_mgr_view = $self->{vim}->get_view(mo_ref => $self->{vim}->get_service_content()->eventManager);
   my $event_collector = $event_mgr_view->CreateCollectorForEvents(filter => $event_filter_spec);
   return $self->{vim}->get_view(mo_ref => $event_collector);
}

sub get_entity_only_events_view {
   my ($self, %args) = @_;
   $self->get_event_collector_view('self', %args);
}

sub get_all_events_view {
   my ($self, %args) = @_;
   $self->get_event_collector_view('all', %args);
}


1;
#########################################################

use 5.006001;
use strict;
use warnings;


#########################################################
package XmlUtil;

sub escape_xml_string {
   my $string = shift;
   if (defined $string) {
      $string =~ s/\&/\&amp;/g;
      $string =~ s/</\&lt;/g;
      $string =~ s/>/\&gt;/g;
   } else {
      $string = undef;
   }
   return $string;
}

sub get_first_child_element {
   my $parent_node = shift;
   my @child_nodes = $parent_node->childNodes;
   my $node;
   foreach $node (@child_nodes) {
      if ($node->nodeType == XML::LibXML::XML_ELEMENT_NODE) {
         return $node;
      }
   }
   return undef;
}

1;
#########################################################



#########################################################
package SoapFault;
use Encode;
use Data::Dumper;
use Class::MethodMaker  [ scalar => [qw(name detail fault_string)] ];

sub new {
   my ($class, $fault_node) = @_;
   my $self = {};
   if (! $fault_node) {
      return bless $self, $class;
   }
   
   my $detail_node = $fault_node->getChildrenByTagName('detail')->shift;
   if (defined($detail_node)) {
      my $fault_detail_child = XmlUtil::get_first_child_element($detail_node);
      my $faultName = $fault_detail_child->localname;
      $faultName =~ /(.*)Fault/;
      $self->{detail} = $1->deserialize($fault_detail_child);
      # bug 194241
      $self->{name} = $faultName;     
   } else {
      my $fault_code_node = $fault_node->getChildrenByTagName('faultcode')->shift;
      $self->{detail} = $fault_code_node->textContent;
   }
   
   my $fault_string_node = $fault_node->getChildrenByTagName('faultstring')->shift;
   $self->{fault_string} = $fault_string_node->textContent;
   return bless $self, $class;
}

use overload '""' => \&stringify;

sub stringify{
   my $self = shift;
   my $fault = "\nSOAP Fault:\n";
   $fault .=   "-----------\n";
   
   if (!Encode::is_utf8($self->{fault_string})) {
      $self->{fault_string} = decode_utf8($self->{fault_string});
   }
   $fault .= "Fault string: " . $self->{fault_string};
   $fault .= "\nFault detail: " . $self->{name};
   return $fault;
}

1;
#########################################################


#########################################################
package SoapResponse;

sub new {
   my ($class, $result, $fault) = @_;
   my $self = { result => $result,
                fault => $fault, };
   return bless $self, $class;
}

sub result {
   my ($self, $val) = @_;
   if ($val) {
      $self->{result} = $val;
      $self;
   } else {
      $self->{result};
   }
}

sub fault {
   my ($self, $val) = @_;
   if ($val) {
      $self->{fault} = $val;
      $self;
   } else {
      $self->{fault};
   }
}

1;
#########################################################



#########################################################
package SoapClient;

use Carp;
use Data::Dumper qw(Dumper);

my $soap_header  = <<'END';
<?xml version="1.0" encoding="UTF-8"?>
   <soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
                     xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
   <soapenv:Body>
END

my $soap_footer = <<'END';
</soapenv:Body></soapenv:Envelope>
END


sub new {
   my ($class, $url) = @_;
   my $user_agent = LWP::UserAgent->new;
   my $cookie_jar = HTTP::Cookies->new(ignore_discard => 1);
   $user_agent->cookie_jar( $cookie_jar );
   $user_agent->protocols_allowed( ['http', 'https'] );
   $user_agent->conn_cache(LWP::ConnCache->new);
   my $self = { user_agent => $user_agent,
                url => $url, };
   return bless $self, $class;   
}

sub save_session {
   my ($self, $file) = @_;
   my $user_agent = $self->{user_agent};
   umask 0077;
   $user_agent->cookie_jar->save($file); 
   return;
}

sub load_session {
   my ($self, $file) = @_;
   my $user_agent = $self->{user_agent};
   $user_agent->cookie_jar->load($file);
   return;
}

sub unload_session {
   my ($self) = @_;
   my $user_agent = $self->{user_agent};
   my $cookie_jar = HTTP::Cookies->new(ignore_discard => 1);
   $user_agent->cookie_jar($cookie_jar);
   return;
}

sub get_session_id {
   my $self = shift;
   my $user_agent = $self->{user_agent};
   $user_agent->cookie_jar->as_string
      =~ m/(.*)vmware_soap_session=\"\\\"([0-9a-zA-Z-](.*)+)\\\"\"(.*)/;
   return $2;
}

sub load_session_id {
   my ($self, $id) = @_;
   my $user_agent = $self->{user_agent};
   my $url = URI::URL::url($self->{url});
   my $host = $url->host();
   $id = '"' . $id . '"';
   my %hash = ();
   $user_agent->cookie_jar->set_cookie(
      0, 'vmware_soap_session', $id, '/',
      $host, undef, 1, undef, undef, 1, \%hash);
}

sub request {
   my ($self, $op_name, $body_content) = @_;
   my $user_agent = $self->{user_agent};
   my $url = $self->{url};

   my $namespace = Vim::get_vim_namespace;
   confess "Not logged in and/or no XML namespace set" unless $namespace;
   my $request_envelope =
      "$soap_header<$op_name xmlns=\"urn:$namespace\">$body_content</$op_name>$soap_footer";

   # http header
   my $http_header = HTTP::Headers->new(
                        Content_Type => 'text/xml',
                        SOAPAction => '"urn:vim25/uber"',
                        Content_Length => byte_length($request_envelope));
   # request
   my $request = HTTP::Request->new('POST',
                                    $url,
                                    $http_header,
                                    $request_envelope);
   
   # send request
   Util::trace(1, "\nREQUEST: " . Dumper($request->content));
   my $response = $user_agent->request($request);
   Util::trace(1, "\nRESPONSE: " . Dumper($response->content) . "\n");
   
   my $xml_parser = XML::LibXML->new;   
   my $result;   
   eval { $result = $xml_parser->parse_string($response->content) };
   if ($@) {
      # response is not well formed xml - possibly be a setup issue
      die "SOAP request error - possibly a protocol issue: " . $response->content . "\n";
   }   
   my $body = $result->documentElement()->getChildrenByTagName('soapenv:Body')->shift;
   my $return_val = $body->getChildrenByTagName("${op_name}Response")->shift;
   if (! $return_val) {
      # must be fault
      $return_val = $body->getChildrenByTagName('soapenv:Fault')->shift;
      if (! $return_val) {
         # neither a valid response or a fault - fatal error
         die "Unexpected response from server: " . $response->content . "\n";         
      }
      # should be trapped by caller
      return (undef, $return_val);
   } else {
      my @returnvals = $return_val->getChildrenByTagName('returnval');
      return (\@returnvals, undef);
   }
}

sub byte_length {
    my ($string) = @_;

    use bytes;
    return length($string);
}

1;
#########################################################


#########################################################
package PrimType;

sub new {
   my ($class, $val, $type_name) = @_;
   my $self = { val => $val , type_name => $type_name};
   return bless $self, $class;   
}

sub serialize {
   my ($self, $tag, $emit_type) = @_;
   my $val = $self->{val};   
   my $serialized_string = "<$tag>";
   if ($emit_type) {
      my $type_name = $self->{type_name};
      $serialized_string = "<$tag xsi:type=\"xsd:${type_name}\">";      
   }
   return $serialized_string .= "$val</$tag>";
}

sub val {
   my $self = shift;
   return $self->{val};
}

1;
#########################################################


#########################################################
package SimpleType;

use Encode;

sub new {
   my ($class, $val) = @_;
   my $self = { val => $val };
   return bless $self, $class;   
}

sub serialize {
   my ($self, $tag, $emit_type) = @_;
   my $val = encode_utf8($self->{val});
   my $serialized_string = "<$tag>";
   if ($emit_type) {
      my $type_name = ref $self;
      $serialized_string = "<$tag xsi:type=\"$type_name\">";      
   }
   return $serialized_string = "$serialized_string$val</$tag>";
}

sub deserialize {
   my ($class, $element_node) = @_;
   if (! defined($element_node)) {
      return undef;
   }
   my $content = $element_node->textContent;
   my $self = { val => decode_utf8($content) };
   return bless $self, $class;
}

sub val {
   my $self = shift;
   return $self->{val};
}

1;
##################################################################################



##################################################################################
package ComplexType;

sub deserialize {
   my ($class, $element_node) = @_;
   if (! defined($element_node)) {
      return undef;
   }
   my $type_node =
      $element_node->getAttributeNodeNS('http://www.w3.org/2001/XMLSchema-instance',
                                        'type');
   if ($type_node) {
      $class = $type_node->textContent;
   }
   my $self = {};
   my @property_list = $class->get_property_list();
   foreach (@property_list) {
      my ($property_name, $class_name, $isarray) = @$_; 
      my @child_nodes = $element_node->getChildrenByTagName($property_name);
      if (! @child_nodes) {
         next;
      }      
      my $property_val = [];   
      foreach (@child_nodes) {
         my $val;
         my $child_class_name = $class_name;
         
         if (defined $child_class_name) {
            my $child_type_node =
               $_->getAttributeNodeNS('http://www.w3.org/2001/XMLSchema-instance', 'type');
            if ($child_type_node) {
               $child_class_name = $child_type_node->textContent;
               if ($child_class_name =~ /^xsd:/) {
                  undef $child_class_name;
               }
            }
         }
         
         if ($child_class_name) {
            if ($child_class_name eq 'boolean') {
               if ($_->textContent eq 'true' or $_->textContent eq '1') {
                  $val = '1';
               } elsif ($_->textContent eq 'false' or $_->textContent eq '0') {
                  $val = '0';
               } else {
                  Carp::confess("Internal error: server returned '$val' as a boolean value");
               }
            } else {
               $val = $child_class_name->deserialize($_);
            }
         } else {
            $val = $_->textContent;
         }
         if ($isarray) {
            $property_val = [@$property_val, $val];
         } else {
            $property_val = $val;
         }
      }      
      my $propValType = ref $property_val;
      if ($propValType =~ /ArrayOf.*/) {
         my @keyvals = %$property_val;
         if (@keyvals) {
            $self->{$property_name} = pop @keyvals;
         }
      } else {
         $self->{$property_name} = $property_val;
      }
   } 
   return bless $self, $class;   
}


sub serialize {
   my ($self, $tag, $emit_type) = @_;
   my @property_list = $self->get_property_list();
   my $serialized_string = "<$tag>";
   if ($emit_type) {
      my $type_name = ref $self;
      $serialized_string = "<$tag xsi:type=\"$type_name\">";
   } 
   foreach (@property_list) {
      my ($property_name, $class_name, $isarray) = @$_;
      
      if (! exists($self->{$property_name})) {
         next;
      }
      my $val = $self->{$property_name};
      my @values;
      if ($isarray) {
         @values = @$val;
      } else {
         @values = $val;
      }
      foreach (@values) {
         if (defined ($class_name)) {
            if ($class_name eq 'boolean') {
               my $val;
               if (lc($_) eq 'true' or $_ eq '1') {
                  $val = '1';
               } elsif (lc($_) eq 'false' or $_ eq '0') {
                  $val = '0';
               } else {
                  Carp::confess "Cannot serialize boolean: acceptable values are true/false/1/0, got '$_'";
               }
               $serialized_string .= "<$property_name>" . XmlUtil::escape_xml_string($val) . "</$property_name>";                          
               next;
            }
            
            # complex type
            my $show_type;
            my $obj_type_name = ref $_;
            if ($class_name ne $obj_type_name) {               
               $show_type = 1;
               if ($class_name eq 'anyType' &&  ! $obj_type_name) {
                  # primitive going out as any -- treat them as xsd:string
                  if (defined ($_)) {
                     $serialized_string .=       
                        "<$property_name xsi:type=\"xsd:string\">" .
                        XmlUtil::escape_xml_string($_) . "</$property_name>";
                  }
                  next;
               }
            }
            if ($obj_type_name && $class_name eq 'ManagedObjectReference') {
               $obj_type_name = 'ManagedObjectReference';   
            }
            if ($class_name ne 'anyType') {
               if (! $obj_type_name || ! $obj_type_name->isa($class_name)) {
                  Carp::confess("Cannot serialize $property_name as $class_name");
               }
            }
            # bug 236635
            if (defined ($_)) {
               $serialized_string .= $_->serialize($property_name, $show_type);
            }
         } else {
            # primitive
            if (defined($_)) {
               $serialized_string .=       
                  "<$property_name>" . XmlUtil::escape_xml_string($_) . "</$property_name>";     
            }         
         }
      }
   }
   return $serialized_string .= "</$tag>";
}


sub new {
   my ($class, %args) = @_;

   # bug 194240
   arg_validation($class, %args);

   my $self = {};
   bless $self, $class;
   if (%args) {
      foreach ($class->get_property_list()) {
         my ($name, $type) = @$_;
         if (exists $args{$name}) {
            $self->{$name} = $args{$name};
         }
      }
   }
   return $self;
}

# bug 194240
sub arg_validation {
   my ($class, %args) = @_;
   my @arg_keys = keys %args;
   my @prop_list = $class->get_property_list();
   foreach(@arg_keys) {
      my $found = 0;
      my $arg = $_;
      foreach my $prop (@prop_list) {
         if($arg eq $prop->[0]) {
            $found = 1;
         }
      } 
      if($found == 0) {
         Carp::confess("Argument $arg is not valid");
      }
   }
}

sub get_property_list {
   return ();
}

1;
##################################################################################


##################################################################################   
package ManagedObjectReference;
our @ISA = qw(ComplexType);

our @property_list = (['type', undef],
                      ['value', undef]);
                      
use Class::MethodMaker  [ scalar => [qw(type value)] ];

sub get_property_list {
   return @property_list;   
}

sub deserialize {
   my ($class, $element_node) = @_;
   if (! defined($element_node)) {
      return undef;
   }
   # deserialize ManagedObjectReference
   my $self = {
      type => $element_node->getAttribute('type'),
      value => $element_node->textContent
   };
   return bless $self, $class;
}

sub serialize {
   my ($self, $tag, $show_type) = @_;
   my $type = $self->{type};
   my $id = $self->{value};
   if ($show_type) {
      return "<$tag xsi:type=\"ManagedObjectReference\" type=\"$type\">$id</$tag>\n";
   } else {
      return "<$tag type=\"$type\">$id</$tag>\n";
   }
}

1;

# bug 158165
END {
   if($is_connected != 0) {
      print "\nEnd Disconnect\n";
      Util::disconnect();
   }
}