#!/usr/bin/perl -w
#
# Copyright 2006 VMware, Inc.  All rights reserved.
#
# This scripts allows users to setup licensing for multiple hosts
#
# Examples of usage
# - assuming the serviceurl, user, password is already setup
#
# license hosts using the license setup file
# hostlicensing.pl --hostop licensing --licfile c:/lic_setup.txt
#
# print out licensing for VC and a single host
# hostlicensing.pl --hostop licenseusage --host myhost.domain
#


use 5.006001;
use strict;
use warnings;

use VMware::VIRuntime;

use AppMainUtil;
use HostUtil;


my $app_opname = 'hostop';

my $lic_usage_opname = 'licenseusage';
my $lic_usage_argname = 'host';

my $licensing_opname = 'licensing';
my $licensing_argname = 'licfile';


my %ops_info = (
   $lic_usage_opname => {
      handler => \&print_license_usage,
      option => $lic_usage_argname,
   },
   $licensing_opname => {
      handler => \&setup_licensing,
      option => $licensing_argname,
   },
);

my %opts = (
   $app_opname => {
      type => "=s",
      help => "Operation to perform on target host: " . 
          join(", ", sort keys %ops_info),
      required => 1,
   },
   $lic_usage_argname => {
      type => "=s",
      help => "Hostname to check licensing for (if managed by VC)",
      required => 0,
   },
   $licensing_argname => {
      type => "=s",
      help => "License Data file with additional information for each host",
      required => 0,
   },
);


my $attr_name = 'name';
my $attr_config_product = 'config.product';

my $all_host_name = 'All';

# edition keys
my $edition_esxStandard = 'esxFull';
my $edition_esxBasic = 'esxBasic';

# feature keys
my $feature_vsmp = 'vsmp';
my $feature_san = 'san';
my $feature_nas = 'nas';
my $feature_iscsi = 'iscsi';
my $feature_backup = 'backup';

my %esx_keys = (
   #edition keys
   $edition_esxStandard => $edition_esxStandard,
   $edition_esxBasic => $edition_esxBasic,
   
   # feature keys
   $feature_vsmp => $feature_vsmp,
   $feature_san => $feature_san,
   $feature_nas => $feature_nas,
   $feature_iscsi => $feature_iscsi,
   $feature_backup => $feature_backup,
);

my $features_standard = [ $feature_vsmp, 
                          $feature_san, 
                          $feature_nas, 
                          $feature_iscsi, 
                        ];
my $features_basic = $feature_nas;

# license edition to feature dependencies
my %edition_features = (
   $edition_esxStandard => {
      feature_list => \$features_standard,
      features => join(',', $features_standard),
      addons => $feature_backup
   },
   $edition_esxBasic => {
      feature_list => \$features_basic,
      features => join(',', $features_basic),
      addons => ''
   }, 
);


# call common application handler with necessary setup
AppUtil::app_setup($app_opname, \%ops_info, \%opts, \&validate);
AppUtil::app_main();


# validate --
#      validates hostop and extra arguments
# returns true on successful validation
sub validate {
   my $valid = 1;

   my $host_op = Opts::get_option('hostop');

   # get handler for operation 
   my $handler = $ops_info{$host_op}{handler};

   if (!defined($handler)) {
      Util::trace(1, "Invalid Operation : " . $host_op);
      $valid = 0;
      return $valid;
   }

   if ($handler == \&setup_licensing) {
      my $lic_file = Opts::get_option($licensing_argname);
      if ((length($lic_file) == 0) || !(-e $lic_file && -r $lic_file && -T $lic_file)) {
         Util::trace(1, "Missing Licensing setup data file : " . $lic_file);
         $valid = 0;
      }
   }
   # we dont validate the 'host' argument. it really is optional.

   return $valid;
}


# print_license_usage --
#      output the license usage for a single host
# vim_service - vim service instance
# sic         - service instance content 
# host        - host managed by VC to check licensing for
sub print_license_usage {
   my ($vim_service, $sic, $host) = @_;

   my $host_ref;
   my $host_type = "ESX Host ";
   if (HostUtil::is_virtual_center($sic->about)) {
      $host_type = "VirtualCenter ";

      # get the moref of the named host
      my %host_info = ();
      my $host_views = Vim::find_entity_views(view_type => 'HostSystem',
                                              filter => {
                                                 $attr_name => $host
                                              },
                                              properties => [
                                                 $attr_name
                                              ]
                                              );

      if (scalar($host_views) == 0) {
         Util::trace(1, "No HostSystems's available to query license usage.\n");
         return;
      }

      # get the host reference from the list.
      my $single_host = @$host_views;
      #$host_ref = $single_host{'mo_ref'};
   }

   # host_moref is undefined if we're logged into ESX
   my $usage_info;
   eval {
      $usage_info = 
         Util::check_fault(
            $vim_service->QueryLicenseUsage(_this => $sic->licenseManager,
                                            host => $host_ref));
   };
   if ($@) {
      Util::trace(0, "Fault Querying usage :" . $@->fault_string);
   }

   my $source_str;
   my $source_type;
   if (defined($usage_info->source)) {
      if ($usage_info->source->isa('LicenseServerSource')) {
         $source_str = $usage_info->source->licenseServer;
         $source_type = 'Served';
      } else {
         $source_type = 'Local File';
      }
   }

   my $server = Opts::get_option('server');
   print "Licensing for $host_type " . $server . "\n";
   print $source_type . " Source " . $source_str . "\n";
   if (defined($host_ref)) {
      print "Including ESX Host : " . $host . "\n";
   }

   my $reservations = $usage_info->reservationInfo;
   foreach (@$reservations) {
      my $reservation = $_;
      my $prefix = "VirtualCenter : " . $server;
      if (defined($host_ref) && defined($esx_keys{ $reservation->key })) {
         $prefix = "ESX host : " . $host;
      }
      print $prefix . " Key : " . $_->key . " State : " . $_->state->val . "\n";
   }
}


# setup_licensing --
#      setup licensing for a single or multiple ESX hosts
# vim_service - vim service instance
# sic         - service instance content 
# lic_file    - the data file for licence config
sub setup_licensing {
   my ($vim_service, $sic, $lic_file) = @_;

   my %lic_setup = ();

   #parse data file into hash. if error, print error message and die
   #validate contents of hash. if error, print error message and die
   get_lic_setup($lic_file, \%lic_setup);

   if (scalar(keys %lic_setup) == 0) {
      Util::trace(1, "No License Setup found.");
      return;
   }

   # setup of licensing for VirtualCenter is not supported
   my $host_views = Vim::find_entity_views(view_type => 'HostSystem',
                                           filter => {
                                              $attr_name => undef,
                                              $attr_config_product => undef
                                           },
                                           properties => [
                                              $attr_name,
                                              $attr_config_product
                                           ]
                                           );

   if (scalar($host_views) == 0) {
      Util::trace(1, "No HostSystems's available to query license usage.\n");
      return;
   }

   # if licensing is not enabled, enable it based on contents of data_file
   foreach (@$host_views) {
      my $host_view = $_;
      check_and_license($vim_service, $sic, $host_view, \%lic_setup);
   }
}


# get_lic_setup --
#      get the license information from license setup file
#      fills in a hash containing license setup keyed by hostname
#
# lic_file - the file containing the lic setup data
# ref_lic_info a reference to license setup info to be filled in
sub get_lic_setup {
   my ($lic_file, $ref_lic_setup) = @_;

   my $LFH;
   open($LFH, $lic_file) or die "Unable to open license setup file : " . $lic_file;

   while(<$LFH>) {
      #skip # specified comments in file
      if ($_ =~ /^(#).*/) {
         next;
      }
      if ($_ !~ /(host|edition|addons|source)/) {
         Util::trace(1, "Invalid license spec: Skipping license info line : " . $_);
         next;
      }
      my %lic_info = 
         $_ =~ /(host|edition|addons|source)=(\S+)*/gm;
      if (length($lic_info{host}) <= 0) {
         Util::trace(1, "Invalid host: Skipping license info line : " . $_);
         next;
      }
      my $lic_source = get_lic_source($lic_info{source});
      if (!defined($lic_source)) {
         Util::trace(1, "Invalid source: Skipping license info line : " . $_);
      } else {
         $lic_info{lic_source} = $lic_source;
         my @requested_addons = split(/,/, $lic_info{addons});
         $lic_info{requested_addons} = \@requested_addons;
         $$ref_lic_setup{$lic_info{host} } = \%lic_info;
      }
   }

   close($LFH);
}


# check_and_license --
#      check and license a single host
# vim_service - vim service instance
# sic - service instance content
# host_info   - a single host moref
# lic_info    - license information for 
sub check_and_license {
   my ($vim_service, $sic, $ref_host_info, $lic_setup) = @_;

   my %host_lic_info = (
      host_info => undef,
      edition => undef,
      features => undef,
      addons => undef,
      source => undef,
      lic_source => undef,
   );

   my $product_info = $ref_host_info->get_property($attr_config_product);
   if (defined($product_info) &&
       HostUtil::is_legacy_host($product_info)) {
      Util::trace(1, "Cannot setup licensing for legacy (pre 3.x) hosts");
      return;
   }

   $host_lic_info{host_info} = $ref_host_info;

   my $host_name = $host_lic_info{host_info}{$attr_name};
   my $default_host_lic_setup = $$lic_setup{$all_host_name};
   my $host_lic_setup = $$lic_setup{$host_name};
   if (!$host_lic_setup && $default_host_lic_setup) {
      $host_lic_setup = $default_host_lic_setup;
   } else {
      Util::trace(1, "No License Setup information present for : " . $host_name);
      return;
   }

   eval {
      get_host_license_usage($vim_service, $sic, \%host_lic_info);
   };
   if ($@) {
      Util::trace(0, "Fault getting license usage for : " . $host_name . " :");
      Util::trace(0, "Fault message : " . $@->fault_string . "\n");
      return;
   }

   get_host_lic_setup($host_lic_setup, \%host_lic_info);

   # now actually setup licensing for the host
   my $source_set = 0;
   my $edition_set = 0;
   my $addons_set = 0;
   # change source if necessary
   if (defined($host_lic_info{lic_source})) {
      eval {
         Util::check_fault(
            $vim_service->ConfigureLicenseSource(_this => $sic->licenseManager,
                                                 host => $ref_host_info->get_property('mo_ref'),
                                                 licenseSource => $host_lic_info{lic_source}));
      };
      if ($@) {
         Util::trace(0, "Fault Configuring source :" . $@->fault_string);
         Util::trace(0, "Host licensing setup incomplete :" . $host_name);
         die $@;
      } else {
         $source_set = 1;
         Util::trace(1, "Configured source for " . $host_name);
      }
   }

   # change edition if necessary
   if (length($host_lic_info{edition}) > 0) {
      eval {
         Util::check_fault(
            $vim_service->SetLicenseEdition(_this => $sic->licenseManager,
                                            host => $ref_host_info->get_property('mo_ref'),
                                            featureKey => $host_lic_info{edition}));
      };
      if ($@) {
         Util::trace(0, "Fault Setting edition : " . $host_lic_info{edition} . " :");
         Util::trace(0, "Fault message : " . $@->fault_string . "\n");
         Util::trace(0, "Host licensing setup incomplete :" . $host_name . "\n");
         die $@;
      } else {
         $edition_set = 1;
         Util::trace(1, "Set Edition $host_lic_info{edition} for " . $host_name . "\n");
      }
   }

   # enable any additional features if necessary
   my $addons = $host_lic_info{addons};
   foreach (@$addons) {
      eval {
         Util::check_fault(
            $vim_service->EnableFeature(_this => $sic->licenseManager,
                                        host => $ref_host_info->get_property('mo_ref'),
                                        featureKey => $_));
      };
      if ($@) {
         Util::trace(0, "Fault Enabling feature : " . $_ . " :");
         Util::trace(0, "Fault message : " . $@->fault_string . "\n");
         Util::trace(0, "Host licensing setup incomplete :" . $host_name . "\n");
         die $@;
      } else {
         $addons_set = 1;
         Util::trace(1, "Enabled Addon $_ for " . $host_name);
      }
   }

   if ($source_set == 0 && $edition_set == 0 && $addons_set == 0) {
      Util::trace(1, $host_name . " was not licensed, since there were no changes.");
   }
}


# get_host_license_usage --
#      query the license usage on the host
#      fills in the host_lic_info passed in
# vim_service - vim service instance
# sic         - service instance content 
# host_lic_info  - hash with no license information for the host
# returns host_lic_info hash containing host license usage
sub get_host_license_usage {
   my ($vim_service, $sic, $ref_host_lic_info) = @_;
   
   # host_moref is undefined if we're logged into ESX
   my $host_ref = $$ref_host_lic_info{host_info}{mo_ref}; 
   my $usage_info;
   eval {
      $usage_info = 
         Util::check_fault(
            $vim_service->QueryLicenseUsage(_this => $sic->licenseManager,
                                            host => $host_ref));
   };
   if ($@) {
      Util::trace(0, "Fault Querying usage :" . $@->fault_string);
      return;
   }
   
   # sourcetype = 0 - none / eval, 1 - served, 2 - local
   if (defined($usage_info) && (defined($usage_info->reservationInfo))) {
      $$ref_host_lic_info{'lic_source'} = $usage_info->source;
      if ($usage_info->source->isa('LicenseServerSource')) {
         $$ref_host_lic_info{'source'} = $usage_info->source->licenseServer;
      }

      # check licensing usage for host for necessary configuration
      my $first = 1;
      my $reservations = $usage_info->reservationInfo;
      foreach (@$reservations) {
         my $reservation = $_;
         # skip all non-esx keys
         if (defined($esx_keys{ $reservation->key }) && 
             ($reservation->state->val eq 'licensed' ||
              $reservation->state->val eq 'unlicensedUse')) {
            if (defined($edition_features{ $reservation->key })) {
               $$ref_host_lic_info{edition} = $reservation->key;
            } else {
               if ($first == 1) { 
                  $$ref_host_lic_info{features} .= $reservation->key;
                  $first = 0;
               } else {
                  $$ref_host_lic_info{features} .= ',' . $reservation->key;
               }
            }
         }
      }
   }
}


# get_host_lic_setup --
#      setup host_lic_info with any licensing changes for the host
#      if no licensing changes are needed, the license info is cleared
#      TODO use the Host's FeatureInfo list to derive any new features
#           to be licensed.
# ref_host_lic_setup - requested host licensing setup
# ref_host_lic_info  - existing lic info to be modified with any deltas
sub get_host_lic_setup {
   my ($ref_host_lic_setup, $ref_host_lic_info) = @_;

   if ($$ref_host_lic_info{source} ne $$ref_host_lic_setup{source}) {
      $$ref_host_lic_info{source} = $$ref_host_lic_setup{source};

      my $lic_source = $$ref_host_lic_setup{lic_source};
      if ($lic_source->isa('LocalLicenseSource') &&
          !defined($$ref_host_lic_setup{lic_source_read})) {
         #read in the entire license file once only when first used
         my $LSH;
         open($LSH, "$lic_source->licenseKeys");
         while (<$LSH>) {
            $lic_source->licenseKeys .= $_;
         }
         $$ref_host_lic_info{lic_source_read} = 1;
      }

      $$ref_host_lic_info{lic_source} = $$ref_host_lic_setup{lic_source};
   } else {
      $$ref_host_lic_info{lic_source} = undef;
      $$ref_host_lic_info{source} = undef;
   }
   
   if ($$ref_host_lic_info{edition} ne $$ref_host_lic_setup{edition}) {
      $$ref_host_lic_info{edition} = $$ref_host_lic_setup{edition};
   } else {
      $$ref_host_lic_info{edition} = undef;
   }

   my $requested_addons = $$ref_host_lic_setup{requested_addons};

   my @add_features;
   foreach (@$requested_addons) {
      if ((defined($$ref_host_lic_info{features}) &&
          $$ref_host_lic_info{features} !~ /^$_$/) ||
          $edition_features{$$ref_host_lic_setup{edition}}{addons} =~ /^$_$/) {
         push @add_features, $_; 
      } else {
         Util::trace(1, "Unknown feature : " . $_);
      }
   }

   $$ref_host_lic_info{addons} = \@add_features;
}


# get_lic_source --
#      checks if the source string is a served source
# source - source string - file path or license file name(.lic)
# returns 0 - no source, 1 - served source, 2 - local source
sub get_lic_source {
   my ($source_str) = @_;

   my $lic_source = 0;
   if (length($source_str) > 0) {
      if (-e $source_str && -r $source_str) {
         # the license key file name is pushed here.
         # it will be pulled, opened and its contents replaced here when needed
         # to configure source
         $lic_source = LocalLicenseSource->new(licenseKeys => $source_str);
      } else {
         #treat everything else as a served source
         #any errors will be handled when configuring source
         $lic_source = LicenseServerSource->new(licenseServer => $source_str);
      }
   }

   return $lic_source;
}


__END__

=head1 NAME

hostslicensing.pl - get license usage or setup licensing for one or more ESX servers

=head1 SYNOPSIS

hostslicensing.pl --hostop [licenseusage | licensing] [options]

=head1 DESCRIPTION

This VI Perl utility has 2 capabilities, 
1. to get the licensing usage for one or more ESX servers including the VirtualCenter server.
2. to setup licensing for one or more ESX servers on a VirtualCenter server.
This utility can also be run for a single standalone ESX 3.x server.
This utility will not setup licensing for legacy 2.x ESX servers or VirtualCenter. 

=head1 OPTIONS

=head2 GENERAL OPTIONS

=over

=item B<hostop>

Required. The operation to be performed on the host. 
licenseusage - get the license usage for the host specified
               in case of a VirtualCenter server, this will print out
               license usage for the VirtualCenter server, and all managed ESX 3.x hosts.  
licensing - setup licensing for one or more ESX hosts, using the license specification
            in a license setup file.


=head2 LICENSEUSAGE OPTIONS

=item B<host>

Optional. The DNS name or IP address of the VC or ESX server to get license usage for.
If the host is specified, the utility will retrieve usage information for that host only.
This parameter is ignored if the utility logs into an ESX 3.x server directly. 


=head2 LICENSING OPTIONS

=item B<licfile>

Required. Path for the text file containing the license setup for one or more hosts.
This file has a very specific format as documented below.
If the license file is empty or unreadable, licensing will not be done.

The license setup file format is key=value pairs separated by whitespace.
Blank lines are skipped.
keys are : host, edition, addons, source, in that order
if a key does not have a value, you must still specify key= 
host - specify a fully qualified domain name or IP address for a specific host
       "All" - settings apply in absense of a specific setting for 
       an individual host
addons - dependend on edition selected 
         vsmp, nas, iscsi, san, backup are allowed addons
edition - one of esxFull, esxExpress, esxBasic, esxVdi
          esxFull - all addons included, "backup" addon used for ESX Enterprise versions.
          esxVdi  - same features and addon as esxFull, and "backup" is included.
          esxBasic - "nas" is the only addon allowed. Excluded - "vsmp", "san", "backup".
          esxExpress - "nas" is the only addon allowed. Excluded - "vsmp", "san", "backup".
source - port@license-server@domain - the port and fully qualified license server name.

The following line licenses all hosts with all features including backup
host=All edition=esxFull addons=backup source=27000@some-lic-server.mydomain.com

The following line makes a special override for a particular host
Any host can have special case licensing setup as in the example below
host=hostname.domain edition=esxExpress addons=san source=port@lic-server.domain


=back


=head1 EXAMPLES

Print license usage for a VirtualCenter Server and all ESX hosts it manages

 hostslicensing.pl --protocol https --server <VC server hostname or ip> 
                   --servicepath /sdk --username <username> --password <passwd> 
                   --hostop licenseusage


Print license usage for a a single host managed by a VirtualCenter Server

 hostslicensing.pl --protocol https --server <VC server hostname or ip> 
                   --servicepath /sdk --username <username> --password <passwd> 
                   --hostop licenseusage --host <ESX 3.x server dns name or ip>


Get Log URLS for all hosts managed by VirtualCenter server to c:\Temp\log_bundles :

 hostslicensing.pl --protocol https --server <VC server hostname or ip> 
                   --servicepath /sdk --username <username> --password <passwd> 
                   --hostop licensing --licfile c:\Temp\lic_setup.txt

 The lic_setup.txt file contains the following contents:

 # The following line licenses all hosts with all features including backup
 host=All edition=esxFull addons=backup source=27000@some-lic-server.mydomain.com

 # The following line makes a special override for a particular host
 # Any host can have special case licensing setup as in the example below
 host=hostname.domain edition=esxExpress addons=san source=port@lic-server.domain
 

=head1 SUPPORTED PLATFORMS

All operations work with VMware VirtualCenter 2.0 or later.

All operations works with VMware ESX Server 3.0 or later.


